//+------------------------------------------------------------------+
//|                                         HedgeAccountCloseAll.mq4 |
//|                                      Copyright 2017, nicholishen |
//|                         https://www.forexfactory.com/nicholishen |
//+------------------------------------------------------------------+
#property copyright "Copyright 2017, nicholishen"
#property link      "https://www.forexfactory.com/nicholishen"
#property version   "1.00"
#property strict

#include <Arrays\ArrayObj.mqh>
#include <stdlib.mqh>
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
class LiveOrder : public CObject
{
protected:
   const string   m_symbol;
public:
   LiveOrder(string symbol):m_symbol(symbol){}
   double Size()const 
   {
      double net_size=0.0;
      for(int i=OrdersTotal()-1;i>=0;i--)
      {
         if(OrderSelect(i,SELECT_BY_POS)&&OrderSymbol()==m_symbol)
         {
            if(OrderType()==OP_BUY)
               net_size+=OrderLots();
            else
            if(OrderType()==OP_SELL)
               net_size-=OrderLots();
         }  
      } 
      return net_size;  
   }
   string Symbol() const  
   { 
      return m_symbol;
   }
   double Risk() const
   {
      return fabs(Size())*SymbolInfoDouble(m_symbol,SYMBOL_TRADE_TICK_VALUE);
   }
   
   virtual int Compare(const CObject *node,const int mode=0)const override
   {
      const LiveOrder* other = node;
      if(this.Risk()>other.Risk())
         return -1;
      if(this.Risk()<other.Risk())
         return 1;
      return 0;
   }
};
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
class PendingOrder : public CObject
{
protected:
   const int m_ticket;
public:
   PendingOrder(int ticket):m_ticket(ticket){}
   int Ticket()
   {
      return m_ticket;
   }
   bool DeletePending()
   {
      if(!OrderSelect(m_ticket,SELECT_BY_TICKET))
         return false;
      if(OrderType()<2)
         return OrderModify(m_ticket,OrderOpenPrice(),0,0,0);
      return OrderDelete(m_ticket);
   }
   int NumTicks()const
   {
      if(!OrderSelect(m_ticket,SELECT_BY_TICKET))
         return 0;
      double tick_size=SymbolInfoDouble(OrderSymbol(),SYMBOL_TRADE_TICK_SIZE);
      if(OrderType()>=2)
         return int(fabs(OrderOpenPrice()-OrderClosePrice())/tick_size);
      double sl = OrderStopLoss() > 0.0 ? OrderStopLoss() : DBL_MAX;
      double tp = OrderTakeProfit() > 0.0 ? OrderTakeProfit() : DBL_MAX;
      return (int)fmin(fabs(OrderOpenPrice()-sl)/tick_size,fabs(OrderOpenPrice()-tp)/tick_size);
   }
   virtual int Compare(const CObject *node,const int mode=0)const override
   {
      const PendingOrder* other = node;
      if(this.NumTicks()>other.NumTicks())
         return 1;
      if(this.NumTicks()<other.NumTicks())
         return -1;
      return 0;
   }
};
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
template<typename T>
class vector:public CArrayObj
{
   public: T operator[](const int i)const{return At(i);}
};
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
class TradeManager
{
   vector<LiveOrder*>      m_live_list;
   vector<PendingOrder*>   m_pending_list;
public:
   void HedgeAll()
   {
      int num_open = 0;
      ulong ms = GetTickCount();
      for(int i=OrdersTotal()-1;i>=0;i--)
      {
         if(OrderSelect(i,SELECT_BY_POS)&&OrderType()<2)
         {
            num_open++;
            bool found = false;
            for(int j=0;j<m_live_list.Total();j++)
            {   
               if(m_live_list[j].Symbol()==OrderSymbol())
               {
                  found=true;
                  break;     
               }
            }
            if(!found)
            {
               m_live_list.Add(new LiveOrder(OrderSymbol()));
            }
         }
      }
      m_live_list.Sort();
      int attempts = 0;
      bool is_hedged = false;
      int num_symbols = m_live_list.Total();
      while(!is_hedged&&!IsStopped()&&attempts<3)
      {
         for(int i=0;i<m_live_list.Total();i++)
         {
            double size = NormalizeDouble(m_live_list[i].Size(),2);
            string symbol = m_live_list[i].Symbol();
            if(size > 0.0)
            {
               double price = SymbolInfoDouble(symbol,SYMBOL_BID);
               if(OrderSend(symbol,OP_SELL,size,price,0,0,0)<0)
                  Print("HedgeOrderError: ",ErrorDescription(GetLastError()));
            }
            else
            if(size < 0.0)
            {
               double price = SymbolInfoDouble(symbol,SYMBOL_ASK);
               if(OrderSend(symbol,OP_BUY,fabs(size),price,0,0,0)<0)
                  Print("HedgeOrderError: ",ErrorDescription(GetLastError()));
            }
         }
         is_hedged=IsHedged();
         attempts++;
      } 
      ms = GetTickCount()-ms;
      if(is_hedged && num_open > 0)
      {
         Alert(string(num_open)+" trades on "+string(num_symbols)+" symbols closed in "+string(ms)+" ms. Order reconciliation will happen after any pending orders are closed.");
      }
   }
   void CloseByHedge()
   {
      for(int s=0;s<m_live_list.Total();s++)
      {
         string symbol = m_live_list[s].Symbol();
         for(int i=OrdersTotal()-1;i>0&&!IsStopped();i--)
         {
            if(OrderSelect(i,SELECT_BY_POS)
               && OrderSymbol()==symbol 
               && OrderType()<2 
               //&& (!m_locked_magic || OrderMagicNumber() == MAGIC)
               )
            {
               int type = OrderType();
               int ticket = OrderTicket();
               for(int j=i-1;OrderSelect(j,SELECT_BY_POS);j--)
               {
                  if(OrderSymbol()==symbol 
                     && OrderType()<2 
                     && OrderType()!=type 
                     //&& (!m_locked_magic || OrderMagicNumber() == MAGIC)
                     )
                  {
                     if(!OrderCloseBy(OrderTicket(),ticket))
                        Print("OrderCLoseByError: ",ErrorDescription(GetLastError()));
                     i=OrdersTotal();
                     break; 
                  }
               }
            }
         }   
      }
   }
   void ClosePending()
   {
      int attempts = 0;
      while(!IsStopped()&&!AllPendingClosed()&&attempts<3)
      {
         for(int i=0;i<m_pending_list.Total();i++)
         {
            if(!m_pending_list[i].DeletePending())
               Print("OrderDeleteError: ",ErrorDescription(GetLastError()));   
         }
         attempts++;
      }
   }
protected:
   bool  IsHedged()
   {
      for(int i=0;i<m_live_list.Total();i++)
         if(NormalizeDouble(m_live_list[i].Size(),2)!=0.0)
            return false;
      return true;
   }
   bool  AllPendingClosed()
   {
      m_pending_list.Clear();
      for(int i=OrdersTotal()-1;i>=0;i--)
      {
         if(OrderSelect(i,SELECT_BY_POS) && (OrderType()>=2 || OrderStopLoss() != 0.0 || OrderTakeProfit() != 0.0))
            m_pending_list.Add(new PendingOrder(OrderTicket()));
      }
      m_pending_list.Sort();
      return m_pending_list.Total() == 0;
   }
};
//+------------------------------------------------------------------+
//| Script program start function                                    |
//+------------------------------------------------------------------+
void OnStart()
{
   TradeManager trades;
   trades.HedgeAll();
   trades.ClosePending();
   trades.CloseByHedge();
}
//+------------------------------------------------------------------+
